DEFAULT_PACKAGE_ENTRY_POINT = 'main.py'
DEFAULT_PACKAGE_CLASS_NAME = 'ServiceRunner'
DEFAULT_PACKAGE_FUNCTION_NAME = 'run'
DEFAULT_PACKAGE_MODULE_NAME = 'default_module'
DEFAULT_PACKAGE_NAME = 'default_package'
